/*
**++
** Copyright (c) 1995-1997 By
** Acorn Software
** All rights reserved.
**
** This software is furnished under a license and may be used and copied
** only in accordance of the terms of such license and with the
** inclusion of the above copyright notice.  This software or any other
** copies thereof may not be provided or otherwise made available to any
** other person.  No title to and ownership of the software is hereby
** transferred.
**
** The information in the software is subject to change without notice
** and should not be construed as a comitment by Acorn Software.
**
** Acorn Software assumes no responsibility for the use or reliability 
** of its software on equipment which is not supplied by Acorn Software.
**
**  FACILITY:
**
**      Virtual Branches
**
**  ABSTRACT:
**
**      Disk arbitration.  Handles scheduling of requests for I/O to virtual
**	disks against physical disks.  For reasons of modularity and information
**	hiding the functions formerly performed by CLUSTRCOMM.FOR have been
**	included here.
**
**  AUTHORS:
**
**      Dick Munroe
**      267 Cox St.
**      Hudson, Ma. 01749
**
**  CREATION DATE:      18-Apr-95
**
**  MODIFICATION HISTORY:
**
**      0.000 Dick Munroe 18-Apr-95
**          Initial Version Created
**
**      0.001 Dick Munroe 24-May-95
**	    Add log and termination for robot failure.
**	    
**	0.002 Dick Munroe 13-Jul-95
**	    Blew the case where multi sided media is present in a multi drive
**	    jukebox.
**	    
**	0.003 Dick Munroe 27-Jul-95
**	    Ignore warnings when setting up locks.  This handles the case of a
**	    node crash when the crashing node holds EX or PW mode locks.
**	    
**	0.004 Dick Munroe 22-Nov-95
**	    Ignore warnings when setting up import/export locks too.
**	    
**	0.005 Dick Munroe 16-Mar-96
**	    Add disable drive mechanism.
**	    
**	0.006 Dick Munroe 26-Dec-96
**	    Get time delays across jpakak calls for debugging purposes.
**	    Spit out OPCOM message if drive fails to come online.
**	    
**	0.007 Dick Munroe 21-Feb-97
**	    Add tracing for startup.
**	    Turns out that the unload all drives routine needed to obey a bit
**	    more of the locking protocol than I originally thought.  It was
**	    wedging the slaves after freeing the first drive.
**	    
**	0.008 Dick Munroe 24-Feb-97
**	    Add to available becomes an external interface.
**	    
**	0.009 Dick Munroe 08-Jul-97
**	    Fix orphan drives bug.
**	    Fix bug in addToAvailable that puts empty drives on the end of the
**	    available list, thus inducing some thrashing.
**	    
**	0.010 Dick Munroe 15-Jul-97
**	    Export Magazine support.
**	    Found what I believe to be is a bug in announceInProgress.  At the
**	    tail end of MasterSlaveIOQuiesced there is logic that asserts that
**	    for [JBUNIT] an announce is in process.  The new jb unit for which
**	    an announce is ocurring is newJBUnit.
**	    
**	0.011 Dick Munroe 01-Aug-97
**	    Select the available drive with the oldest active I/O time.
**	    
**	0.012 Dick Munroe 05-Sep-97
**	    Check for a JB unit actually IN the drive in cancel master dwell.
**	    
**	[@tbs@]...
**--
*/

#if defined(__ALPHA)
#pragma nomember_alignment
#endif

/*
**
**  INCLUDE FILES
**
*/

#include "brnch-acp-data-structures.h"
#include "sizing.h"
#include "sqsdef.h"

#include <descrip.h>
#include <lckdef.h>
#include <lib$routines.h>
#include <libdef.h>
#include <libdtdef.h>
#include <lkidef.h>
#include <rab.h>
#include <ssdef.h>
#include <starlet.h>
#include <stdio.h>
#include <stdlib.h>
#include <stsdef.h>
#include <time.h>

#ifndef LCK$M_QUECVT
#define LCK$M_QUECVT 0x00001000
#endif

/*
** FIX THIS: This shouldn't be keyed off of jb unit numbers.  used the
** information in the data structures.
*/

#define brnch_isFlipNecessary(a, b) (a != b) && ((a & -2) == (b & -2)) && (acpgeo.brnch$l_numSides != 1) ;

static void				addToBusy(unsigned long) ;
extern void				brnch_addToAvailable(unsigned long) ;
void					brnch_addToPending(unsigned long) ;
extern unsigned long			brnch_amIMaster() ;
void					brnch__cancelMasterDwell(unsigned long) ;
extern unsigned long			brnch__closeIT() ;
extern unsigned long			brnch_connectJBUnitToDrive(unsigned long, unsigned long, unsigned long) ;
void					brnch_diskArbitration(brnch$r_sndcmd_message*) ;
extern unsigned long			brnch_dispatchRoutine(void (*)(unsigned long), unsigned long) ;
extern unsigned long			brnch_driveToVolume(unsigned long, unsigned long) ;
extern unsigned long			brnch_flipInDrive(unsigned long) ;
unsigned long				brnch_isTraceArbitrationOn() ;
unsigned long				brnch_isTraceMiscellaneousOn() ;
extern unsigned long			brnch__logit(char*, ...) ;
extern unsigned long			brnch__openIT(unsigned long) ;
extern unsigned long			brnch__readIT(struct brnch$r_inventory_table*, unsigned long) ;
unsigned long				brnch_queueIOPost(unsigned long jbUnit) ;
void					brnch_removeFromAvailable(unsigned long) ;
unsigned long				brnch_removeFromPending(unsigned long) ;
extern unsigned long			brnch_sendCommandMessage (unsigned long, ...) ;
void					brnch_setJBUnitDwell(unsigned long, unsigned long) ;
void					brnch_setJBUnitPriority(unsigned long, unsigned long) ;
unsigned long				brnch_setUpLocks() ;
unsigned long				brnch_tearDownLocks() ;
unsigned long				brnch_unloadAllDrives() ;
extern unsigned long			brnch_unloadPhysicalDevice(unsigned long) ;
extern unsigned long			brnch__updateIT(struct brnch$r_inventory_table*, unsigned long) ;
extern unsigned long			brnch_volumeToDrive(unsigned long, unsigned long) ;
extern unsigned long			gatim() ;
extern unsigned long			guflg(unsigned long) ;
extern unsigned long			ibsyck(unsigned long) ;
extern unsigned long			jpakak(unsigned long, unsigned long) ;
static void				masterAnnounceGrantedAST(unsigned long) ;
static void				masterDiskArbitration(brnch$r_sndcmd_message*) ;
static void				masterDwellExpired(unsigned long) ;
static void				masterDwellExpiredAST(brnch$r_jbUnit*) ;
static void				masterIOPost(brnch$r_sndcmd_message*) ;
static void				masterPeriodicIOStatusBlockingAST(unsigned long) ;
static void				masterProcessPendingList(unsigned long) ;
static void				masterQuiesceSlaveIOGrantedAST(unsigned long) ;
static void				masterSetDwell(unsigned long) ;
static void				masterSlaveIOQuiesced(unsigned long) ;
static void				masterSlaveIOQuiescedGrantedAST(unsigned long) ;
static void				masterStartIO(brnch$r_sndcmd_message*) ;
static void				masterStartIOBlockingAST(unsigned long) ;
static void				printDriveList(char*, long) ;
static void				printJBUnitList(char*, long) ;
static unsigned long			removeFromBusy(unsigned long) ;
static void				requeueJBUnits(unsigned long) ;
static int				selectLRUAvailableDrive() ;
static void				slaveAnnounce(brnch$r_announce_message*) ;
static void				slaveAnnounceBlockingAST(unsigned long) ;
static void				slaveAnnounceGrantedAST(unsigned long) ;
static void				slaveDiskArbitration(brnch$r_sndcmd_message*) ;
static void				slaveIOPost(brnch$r_sndcmd_message*) ;
static void				slaveQuiesce(brnch$r_sndcmd_message*) ;
static void				slaveQuiesceJBUnitBlockingAST(unsigned long) ;
static void				slaveStartIO(brnch$r_sndcmd_message*) ;
static void				slaveStartIOGrantedAST(unsigned long) ;
static unsigned long			turnOnPhysicalDrive(unsigned long, unsigned long) ;
static void				turnOnPhysicalDriveFailed(unsigned long, unsigned long) ;

static brnch$r_announceLSB		*announceLSB = NULL ;
static brnch$r_jbUnitQuiescedLSB	*jbUnitQuiescedLSB = NULL ;
static brnch$r_periodicIOStatusLSB	periodicIOStatusLSB ;
static brnch$r_quiesceJBUnitLSB		*quiesceJBUnitLSB = NULL ;
static brnch$r_startIOLSB		*startIOLSB = NULL ;
							
static unsigned long			ioPostInProgress = 0 ;

extern struct brnch$r_configuration_entry cte ;		/* Equivalent to cte common block.					*/
extern long				brnch_gl_availableDriveList[2] ;
extern long				brnch_gl_busyDriveList[2] ;
extern long				brnch_gl_dwell ;
extern long				brnch_gl_pendingJBUnits[2] ;


/*
** Add a physical drive to the available list.  Under normal circumstances, the
** drive is always added to the last element on the list but if we add an empty
** drive (which can happen in the event of an export) then we need to add it to
** the FRONT of the available list.
*/

extern void
brnch_addToAvailable(
    unsigned long drive)
{							
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: brnch_addToAvailable(%d)\n", drive) ;
	printDriveList("    list", brnch_gl_availableDriveList[0]) ;
    }
    
    acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onAvailable = 1; 

    if (brnch_gl_availableDriveList[0] == -1)
    {
	/*
	** There aren't any drives on the list.
	*/
	
	acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	brnch_gl_availableDriveList[0] = brnch_gl_availableDriveList[1] = drive ;
    }
    else
    {
	if (acpgeo.brnch$a_drives[drive].brnch$l_jbUnit != -1)
	{
	    /*
	    ** Under normal circumstances, the newly available drive should go at
	    ** the END of the available list, making it the last drive to be chosen.
	    */
	    
	    acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	    brnch_gl_availableDriveList[1] = acpgeo.brnch$a_drives[brnch_gl_availableDriveList[1]].brnch$l_available = drive ;
	}
	else
	{
	    /*
	    ** Avoid thrashing by putting empty, available drives on the FRONT of
	    ** the available list.
	    */

	    acpgeo.brnch$a_drives[drive].brnch$l_available = brnch_gl_availableDriveList[0] ;
	    brnch_gl_availableDriveList[0] = drive ;
	}
    }
        
    if (brnch_isTraceArbitrationOn())
    {
	printDriveList("    list", brnch_gl_availableDriveList[0]) ;
	brnch__logit("Exiting: brnch_addToAvailable(%d)\n", drive) ;
    }
    return ;
} ;

extern void
brnch_addToPending(
    unsigned long jbUnit)
{							
    int					lastJBUnit ;
    int					thisJBUnit ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: brnch_addToPending(%d)\n", jbUnit) ;
	printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
    }

    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onPending = 1; 

    if (brnch_gl_pendingJBUnits[0] == -1)
    {
	/*
	** There aren't any jb units on the list.
	*/
	
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = -1 ;
	brnch_gl_pendingJBUnits[0] = brnch_gl_pendingJBUnits[1] = jbUnit ;
	if (brnch_isTraceArbitrationOn())
	{
	    printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
	    brnch__logit("Exiting: brnch_addToPending(%d)\n", jbUnit) ;
	}
	return ;
    }

    /*
    ** Check for specific case of adding to the BACK of the list.
    */
    
    if (acpgeo.brnch$a_jbUnits[brnch_gl_pendingJBUnits[1]].brnch$l_priority >=
        acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_priority)
    {
	/*
	** The priority of the jb unit being added is LOWER than the current last
	** entry.  Thus we add it at the low end of the list thus maintaining
	** priority AND lru order.
	*/
	
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = -1 ;
	brnch_gl_pendingJBUnits[1] = acpgeo.brnch$a_jbUnits[brnch_gl_pendingJBUnits[1]].brnch$l_pending = jbUnit ;
	if (brnch_isTraceArbitrationOn())
	{
	    printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
	    brnch__logit("Exiting: brnch_addToPending(%d)\n", jbUnit) ;
	}
	return ;
    }
    
    /*
    ** Check for specific case of adding to the FRONT of the list.
    */
    
    if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_priority >
        acpgeo.brnch$a_jbUnits[brnch_gl_pendingJBUnits[0]].brnch$l_priority)
    {
	/*
	** The priority of the jb unit being added is HIGHER than the current first
	** entry.  Thus we add it at the high end of the list maintaining
	** priority AND lru order.
	*/
	
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = brnch_gl_pendingJBUnits[0] ;
	brnch_gl_pendingJBUnits[0] = jbUnit ;
	if (brnch_isTraceArbitrationOn())
	{
	    printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
	    brnch__logit("Exiting: brnch_addToPending(%d)\n", jbUnit) ;
	}
	return ;
    }
    
    /*
    ** Since we have already checked for addition at the beginning and end of
    ** the list, we are guaranteed that an insertion must occur somewhere in the
    ** middle.
    */
    
    thisJBUnit = brnch_gl_pendingJBUnits[0] ;
    
    for (;;)
    {
	lastJBUnit = thisJBUnit ;
	thisJBUnit = acpgeo.brnch$a_jbUnits[thisJBUnit].brnch$l_pending;
	if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_priority >
	    acpgeo.brnch$a_jbUnits[thisJBUnit].brnch$l_priority)
	{
	    /*
	    ** The priority of the new jb unit is greater than the priority of
	    ** the jb unit pointed to by thisJBUnit.  Add the news jb unit
	    ** before thisJBUnit.
	    */
	    
	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = thisJBUnit ;
	    acpgeo.brnch$a_jbUnits[lastJBUnit].brnch$l_pending = jbUnit ;
	    if (brnch_isTraceArbitrationOn())
	    {
		printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
		brnch__logit("Exiting: brnch_addToPending(%d)\n", jbUnit) ;
	    }
	    return ;
	}
    }
} ;

void
brnch__cancelMasterDwell(
    unsigned long drive)
{
    if (drive >= acpgeo.brnch$l_numDrives)
    {
	static $DESCRIPTOR(location, "void brnch__cancelMasterDwell():0") ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    /*
    ** Cancel the master dwell timer for the JB unit currently in the physical
    ** drive.  Of course, this should ONLY be done if there actually IS a JB
    ** unit in the physical drive.
    */

    if (acpgeo.brnch$a_drives[drive].brnch$l_jbUnit != -1)
    {
	sys$cantim (&acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[drive].brnch$l_jbUnit], 0) ;
	sys$dclast (masterDwellExpiredAST, &acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[drive].brnch$l_jbUnit], 0) ;
    }
    else
    {
	brnch__logit("Exiting: void brnch__cancelMasterDwell(%d) had -1 for a jbUnit\n", drive) ;
    }
} ;

/*
** The disk arbitration algorithm is a priority based algorithm.  JB Units with
** higher priority numbers get immediate access to drives.  Within a priority,
** arbitration is strictly LRU with the OLDEST pending I/O request having
** priority.
**
** The module itself is a series of cooperating state machines connected via the
** event queuing mechanism buried in ASNJB.
*/

extern void
brnch_diskArbitration(
    brnch$r_sndcmd_message* message)
{
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Message %p\n", message) ;
	brnch__logit("    irp =         %08x\n", message -> brnch$a_irp) ;
	brnch__logit("    unitNumber =  %08x\n", message -> brnch$l_unit_number) ;
	brnch__logit("    ucb =         %08x\n", message -> brnch$a_ucb) ;
	brnch__logit("    ccb =         %08x\n", message -> brnch$a_ccb) ;
	brnch__logit("    function =    %08x\n", message -> brnch$l_function) ;
	brnch__logit("    timestamp =   %08x\n", message -> brnch$l_timestamp) ;
    }

    if (brnch_amIMaster())
    {
	masterDiskArbitration(message) ;
    }
    else
    {
	slaveDiskArbitration(message) ;
    }
} ;

/*
** On the SLAVE side, only those startIOs that originate on the slave are
** actually recorded in the pending JB unit list.  This means that the slave can
** see "unexpected" announcements that a drive is loaded.  On the SLAVE side, we
** have to be able to ignore the case that occurs when a jb unit ISN'T on the
** pending list.
*/

unsigned long
brnch_removeFromPending (
    unsigned long jbUnit)
{
    int					lastJBUnit ;
    int					thisJBUnit ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: brnch_removeFromPending(%d)\n", jbUnit) ;
	printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
    }
    
    if (brnch_gl_pendingJBUnits[0] == -1)
    {
	return 0 ;
    }

    if (brnch_gl_pendingJBUnits[0] == jbUnit)
    {
	if (brnch_gl_pendingJBUnits[1] == jbUnit)
	{
	    /*
	    ** This is the only element on the list, so reset the list to
	    ** indicate that it is empty.  For drill, make sure the pending
	    ** pointer is cleared as well.
	    */
	    
	    brnch_gl_pendingJBUnits[0] = brnch_gl_pendingJBUnits[1] = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = -1 ;
	}
	else
	{
	    /*
	    ** This is the first element on the list, so remove it from the
	    ** list.  For drill, make sure the pending pointer is cleared as
	    ** well.
	    */
	    
	    brnch_gl_pendingJBUnits[0] = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending ;
	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = -1 ;
	}
	if (brnch_isTraceArbitrationOn())
	{
	    printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
	    brnch__logit("Exiting: brnch_removeFromPending(%d)\n", jbUnit) ;
	}
	
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onPending = 0 ;
	return 1 ;
    } ;

    /*
    ** The requested drive isn't the first one on the list, so we have to look
    ** for it.
    */

    thisJBUnit = brnch_gl_pendingJBUnits[0] ;

    for (;;)
    {
	lastJBUnit = thisJBUnit ;
	thisJBUnit = acpgeo.brnch$a_jbUnits[thisJBUnit].brnch$l_pending ;
	if (thisJBUnit == -1)
	{
	    if (brnch_isTraceArbitrationOn())
	    {
		printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
		brnch__logit("Exiting: brnch_removeFromPending(%d)\n", jbUnit) ;
	    }
	    return 0 ;
	}

	if (thisJBUnit == jbUnit)
	{
	    acpgeo.brnch$a_jbUnits[lastJBUnit].brnch$l_pending = acpgeo.brnch$a_jbUnits[thisJBUnit].brnch$l_pending ;
	    if (acpgeo.brnch$a_jbUnits[lastJBUnit].brnch$l_pending == -1)
	    {
		brnch_gl_pendingJBUnits[1] = lastJBUnit ;
	    } ;
	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_pending = -1 ;
	    if (brnch_isTraceArbitrationOn())
	    {
		printJBUnitList("    list", brnch_gl_pendingJBUnits[0]) ;
		brnch__logit("Exiting: brnch_removeFromPending(%d)\n", jbUnit) ;
	    }
	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onPending = 0 ;
	    return 1 ;
	}
    }
} ;

extern void
brnch_setJBUnitDwell (
    unsigned long jbUnit,
    unsigned long dwell)
{
    unsigned long			i ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: brnch_setJBUnitDwell(%d, %d)\n", jbUnit, dwell) ;
    }

    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_dwell = dwell ;
    for (i = 0; i < acpgeo.brnch$l_numDrives; i++)
    {
	if (acpgeo.brnch$a_drives[i].brnch$l_jbUnit != -1)
	{
	    /*
	    ** The dwell time changed for a JB unit.  Rather than go to a
	    ** whole lot of trouble for something that basically occurs
	    ** infrequently, simply cancel all dwell times.  Then let the
	    ** normal mechanisms propagate the dwell time the NEXT time the
	    ** unit is put into a drive.
	    */
	    
	    brnch__cancelMasterDwell(i) ;
	}
    }
    
    if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive != -1)
    {
    }

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: brnch_setJBUnitDwell(%d, %d)\n", jbUnit, dwell) ;
    }
    return ;
}

extern void
brnch_setJBUnitPriority (
    unsigned long jbUnit,
    unsigned long priority)
{
    unsigned long			drive ;
    unsigned long			i ;
    unsigned long			status ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: brnch_setJBUnitPriority(%d, %d)\n", jbUnit, priority) ;
    }

    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_priority = priority ;
    if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onPending)
    {
	brnch_removeFromPending(jbUnit) ;
	brnch_addToPending(jbUnit) ;
    }

    for (i = 0; i < acpgeo.brnch$l_numDrives;  i++)
    {
	if (acpgeo.brnch$a_drives[i].brnch$l_jbUnit != -1)
	{
	    /*
	    ** The priority changed for a JB unit.  This means that, if
	    ** the jb unit is in a drive and the drive is on the busy list,
	    ** that the drive busy list is now potentially in an
	    ** inconsistent state.  Rather than go to a whole lot of
	    ** trouble for something that basically occurs infrequently,
	    ** simply cancel the dwell time.  Then let the normal
	    ** mechanisms propagate the priority the NEXT time the unit is
	    ** put into a drive.
	    */
	    
	    brnch__cancelMasterDwell(i) ;
	}
    }

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: brnch_setJBUnitPriority(%d, %d)\n", jbUnit, priority) ;
    }
    return ;
}

extern unsigned
long brnch_setUpLocks ()
{
    unsigned long			i ;
    char				lockName[32] ;
    struct dsc$descriptor_s		lockNameDesc = {0, DSC$K_DTYPE_T, DSC$K_CLASS_S, lockName} ;
    unsigned long			status ;

    if (brnch_isTraceMiscellaneousOn())
    {
    	brnch__logit(
	    "extern unsigned long brnch_setUpLocks ()\n") ;
    }

    startIOLSB = (brnch$r_startIOLSB*) calloc (acpgeo.brnch$l_numSlots, sizeof(brnch$r_startIOLSB)) ;
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	startIOLSB[i].brnch$l_status = 0 ;
	startIOLSB[i].brnch$l_lockId = 0 ;
	startIOLSB[i].brnch$l_jbUnit = i ;
	startIOLSB[i].brnch$l_count = 0 ;
    }

    quiesceJBUnitLSB = (brnch$r_quiesceJBUnitLSB*) calloc (acpgeo.brnch$l_numSlots, sizeof(brnch$r_quiesceJBUnitLSB)) ;
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	quiesceJBUnitLSB[i].brnch$l_status = 0 ;
	quiesceJBUnitLSB[i].brnch$l_lockId = 0 ;
	quiesceJBUnitLSB[i].brnch$l_jbUnit = i ;
	quiesceJBUnitLSB[i].brnch$l_count = 0 ;
    }

    jbUnitQuiescedLSB = (brnch$r_jbUnitQuiescedLSB*) calloc (acpgeo.brnch$l_numSlots, sizeof(brnch$r_jbUnitQuiescedLSB)) ;
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	jbUnitQuiescedLSB[i].brnch$l_status = 0 ;
	jbUnitQuiescedLSB[i].brnch$l_lockId = 0 ;
	jbUnitQuiescedLSB[i].brnch$l_jbUnit = i ;
	jbUnitQuiescedLSB[i].brnch$l_count = 0 ;
    }

    memset (&periodicIOStatusLSB, 0, sizeof(periodicIOStatusLSB)) ;

    announceLSB = (brnch$r_announceLSB*) calloc(acpgeo.brnch$l_numDrives, sizeof(brnch$r_announceLSB)) ;
    for (i = 0;  i < acpgeo.brnch$l_numDrives;  i++)
    {
	announceLSB[i].brnch$l_status = 0 ;
	announceLSB[i].brnch$l_lockId = 0 ;
	announceLSB[i].brnch$l_drive = i ;
	announceLSB[i].brnch$l_jbUnit = 0 ;
    }

    /*
    ** Set up the periodic I/O status lock.  The master accumulates all
    ** statistical information about drives.  All slaves ignore them.
    */
    
    sprintf (lockName, "VB$PER%c", cte.brnch$t_jbDevice.dsc$a_pointer[2]) ;
    lockNameDesc.dsc$w_length = strlen (lockName) ;
    if (brnch_amIMaster())
    {
	status = sys$enqw (0, LCK$K_CRMODE, &periodicIOStatusLSB, (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc,
			   0, 0, 0, masterPeriodicIOStatusBlockingAST, 0, 0) ;
    }
    else
    {
	status = sys$enqw (0, LCK$K_NLMODE, &periodicIOStatusLSB, (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc,
			   0, 0, 0, 0, 0, 0) ;
    }
    if ($VMS_STATUS_SUCCESS(status))
    {
	status = periodicIOStatusLSB.brnch$l_status ;
    }
    if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
    {
	static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():0") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    /*
    ** Set up drive loaded locks.
    */
    
    for (i = 0;  i < acpgeo.brnch$l_numDrives;  i++)
    {
	sprintf (lockName, "VB$DRIVE%c%d", cte.brnch$t_jbDevice.dsc$a_pointer[2], i) ;
	lockNameDesc.dsc$w_length = strlen (lockName) ;
	if (brnch_amIMaster())
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &announceLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_NODLCKWT), &lockNameDesc, 
			       0, 0, 0, 0, 0, 0) ;
	}
	else
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &announceLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_NODLCKWT), &lockNameDesc, 
			       0, 0, i, slaveAnnounceBlockingAST, 0, 0) ;
	}
	if ($VMS_STATUS_SUCCESS(status))
	{
	    status = announceLSB[i].brnch$l_status ;
	}
	if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
	{
	    static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():1") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
	}
    }

    /*
    ** Set up startio device locks.
    */
    
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	sprintf (lockName, "VB$STARTIO%c%d", cte.brnch$t_jbDevice.dsc$a_pointer[2], i) ;
	lockNameDesc.dsc$w_length = strlen (lockName) ;
	if (brnch_amIMaster())
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &startIOLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, i, masterStartIOBlockingAST, 0, 0) ;
	}
	else
	{
	    status = sys$enqw (0, LCK$K_NLMODE, &startIOLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, 0, 0, 0, 0) ;
	}
	if ($VMS_STATUS_SUCCESS(status))
	{
	    status = startIOLSB[i].brnch$l_status ;
	}
	if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
	{
	    static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():2") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
	}
    }

    /*
    ** Set up request to quiesce device locks.
    */
    
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	sprintf (lockName, "VB$QUIESCE%c%d", cte.brnch$t_jbDevice.dsc$a_pointer[2], i) ;
	lockNameDesc.dsc$w_length = strlen (lockName) ;
	if (brnch_amIMaster())
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &quiesceJBUnitLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, 0, 0, 0, 0) ;
	}
	else
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &quiesceJBUnitLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, i, slaveQuiesceJBUnitBlockingAST, 0, 0) ;
	}
	if ($VMS_STATUS_SUCCESS(status))
	{
	    status = quiesceJBUnitLSB[i].brnch$l_status ;
	}
	if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
	{
	    static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():3") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
	}
    }

    /*
    ** Set up quiesce device acknowledgement locks.
    */
    
    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	sprintf (lockName, "VB$QUIESCED%c%d", cte.brnch$t_jbDevice.dsc$a_pointer[2], i) ;
	lockNameDesc.dsc$w_length = strlen (lockName) ;
	if (brnch_amIMaster())
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &jbUnitQuiescedLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, 0, 0, 0, 0) ;
	}
	else
	{
	    status = sys$enqw (0, LCK$K_CRMODE, &jbUnitQuiescedLSB[i], (LCK$M_SYSTEM+LCK$M_VALBLK), &lockNameDesc, 
			       0, 0, 0, 0, 0, 0) ;
	}
	if ($VMS_STATUS_SUCCESS(status))
	{
	    status = jbUnitQuiescedLSB[i].brnch$l_status ;
	}
	if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
	{
	    static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():4") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
	}
    }

    /*
    ** Set up the import/export locks.
    */
    
    status = brnch_setupImportExportLocks(&cte.brnch$t_name) ;
    if ((!$VMS_STATUS_SUCCESS(status)) && ($VMS_STATUS_SEVERITY(status) != STS$K_WARNING))
    {
	static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():5") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    /*
    ** Set up master communication lock.
    */
    
    status = brnch_setupMasterComm(&cte.brnch$t_name) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location, "extern unsigned long brnch_setUpLocks ():6") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    if (brnch_isTraceMiscellaneousOn())
    {
    	brnch__logit(
	    "%08x = extern unsigned long brnch_setUpLocks ()\n",
	    status) ;
    }

    return status ;
}

extern unsigned long
brnch_tearDownLocks ()
{
    unsigned long			i ;
    unsigned long			status ;

    if (brnch_isTraceMiscellaneousOn())
    {
    	brnch__logit(
	    "extern unsigned long brnch_tearDownLocks ()\n") ;
    }

    status = sys$deq (periodicIOStatusLSB.brnch$l_lockId,0,0,0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():0") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR,1, &location) ;
    }

    for (i = 0;  i < acpgeo.brnch$l_numDrives;  i++)
    {
	status = sys$deq(announceLSB[i].brnch$l_lockId, 0, 0, 0) ;
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():1") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR,1, &location) ;
	}
    }

    for (i = 0;  i < acpgeo.brnch$l_numSlots;  i++)
    {
	status = sys$deq(startIOLSB[i].brnch$l_lockId, 0, 0, 0) ;
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():2") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR,1, &location) ;
	}

	status = sys$deq(quiesceJBUnitLSB[i].brnch$l_lockId, 0, 0, 0) ;
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():3") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR,1, &location) ;
	}

	status = sys$deq(jbUnitQuiescedLSB[i].brnch$l_lockId, 0, 0, 0) ;
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():4") ;
	    lib$signal (status, 0) ;
	    lib$signal (BRNCH$_INTLOGERR,1, &location) ;
	}
    }

    status = brnch_tearDownImportExportLocks() ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():5") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR,1, &location) ;
    }

    status = brnch_tearDownMasterComm() ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location,"extern unsigned long brnch_tearDownLocks ():6") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR,1, &location) ;
    }

    if (startIOLSB != NULL)
    {
	cfree (startIOLSB) ;
    }

    if (quiesceJBUnitLSB != NULL)
    {
	cfree (quiesceJBUnitLSB) ;
    }

    if (jbUnitQuiescedLSB != NULL)
    {
	cfree (jbUnitQuiescedLSB) ;
    }

    if (announceLSB != NULL)
    {
	cfree (announceLSB) ;
    }

    if (brnch_isTraceMiscellaneousOn())
    {
    	brnch__logit(
	    "%08x = extern unsigned long brnch_tearDownLocks ()\n",
	    status) ;
    }

    return status ;
}

/*
**  This routine unloads all the physical drives.  It's called only by the master
**  ACP at shutdown.  It also updates the database when the platter has been
**  removed.
*/

extern unsigned long
brnch_unloadAllDrives ()
{
    int					drive ;
    int					jbUnit ;
    int					status = SS$_NORMAL ;
    
    for (drive = 0;  drive < acpgeo.brnch$l_numDrives;  drive++)
    {
	if ((jbUnit = acpgeo.brnch$a_drives[drive].brnch$l_jbUnit) != -1)
	{

	    status = sys$enq (0, LCK$K_EXMODE, &jbUnitQuiescedLSB[jbUnit],
			      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
			      0, 0, 0, 0, 0, 0) ;
							/* Arm notification from slaves of successful quiescence of the request	*/
							/* JB unit.								*/
	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		$DESCRIPTOR(xxx,"extern unsigned long brnch_unloadAllDrives ():0") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }

	    status = sys$enqw (0, LCK$K_EXMODE, &quiesceJBUnitLSB[jbUnit],
			       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
			       0, 0, 0, 0, 0, 0) ;
							/* Tell the slaves to quiesce I/O on this JB unit.			*/
	    if ($VMS_STATUS_SUCCESS(status))
	    {
		status = quiesceJBUnitLSB[jbUnit].brnch$l_status ;
	    }

	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		$DESCRIPTOR(xxx,"extern unsigned long brnch_unloadAllDrives ():1") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }

	    sys$synch (0, &jbUnitQuiescedLSB[jbUnit]) ;	/* Wait for the slaves to quiesce the I/O on this JBunit.  Since we're	*/
							/* terminating the acp, we don't have to obey the lock discipline.	*/
	    /*
	    ** Release the arming lock to let all the slaves continue to let
	    ** them move onto more drives.
	    */
	    
	    sys$enqw (0, LCK$K_CRMODE, &jbUnitQuiescedLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		      0, 0, 0, 0, 0, 0) ;

	    status = jbUnitQuiescedLSB[jbUnit].brnch$l_status ;

	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		$DESCRIPTOR(xxx,"extern unsigned long brnch_unloadAllDrives ():2") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }

	    brnch_unloadPhysicalDevice(drive) ;

	    if (acpgeo.brnch$l_pioneer604x == 0)
	    {
		status = brnch_driveToVolume(drive, jbUnit) ;
	    }
	    else
	    {
		status = SS$_NORMAL ;
	    }

	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		brnch__logit(
		    "Failed to unload volume %d (slot %d) from drive %d (%0.*s)",
		    jbUnit,
		    jbUnit/acpgeo.brnch$l_numSides,
		    drive,
		    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$w_length,
		    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$a_pointer) ;
		lib$signal (status, 0) ;
	    }
	    else
	    {
		struct brnch$r_inventory_table	diskRecord ;

		status = brnch__openIT(RAB$C_KEY) ;
		status = brnch__readIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS);
		diskRecord.brnch$r_overlay.brnch$l_contents[drive] = -1 ;
		status = brnch__updateIT(&diskRecord, diskRecord.brnch$l_iunit) ;
		status = brnch__closeIT();
	    }	    
	}
    }

    return status ;
}

static void
addToBusy(
    unsigned long drive)
{							
    int					lastDrive ;
    int					thisDrive ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: addToBusy(%d)\n", drive) ;
	printDriveList("    list", brnch_gl_busyDriveList[0]) ;
    }
    
    acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onBusy = 1; 

    if (brnch_gl_busyDriveList[0] == -1)
    {
	/*
	** There aren't any drives on the list.
	*/
	
	acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	brnch_gl_busyDriveList[0] = brnch_gl_busyDriveList[1] = drive ;
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_busyDriveList[0]) ;
	    brnch__logit("Exiting: addToBusy(%d)\n", drive) ;
	}
	return ;
    }

    /*
    ** Check for specific case of adding to the BACK of the list.
    */
    
    if (acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[brnch_gl_busyDriveList[1]].brnch$l_jbUnit].brnch$l_priority >
        acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[drive].brnch$l_jbUnit].brnch$l_priority)
    {
	/*
	** The priority of the drive being added is LOWER than the current last
	** entry.  Thus we add it at the low end of the list thus maintaining
	** priority AND lru order.
	*/
	
	acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	brnch_gl_busyDriveList[1] = acpgeo.brnch$a_drives[brnch_gl_busyDriveList[1]].brnch$l_available = drive ;
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_busyDriveList[0]) ;
	    brnch__logit("Exiting: addToBusy(%d)\n", drive) ;
	}
	return ;
    }
    
    /*
    ** Check for specific case of adding to the FRONT of the list.
    */
    
    if (acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[drive].brnch$l_jbUnit].brnch$l_priority >=
        acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[brnch_gl_busyDriveList[0]].brnch$l_jbUnit].brnch$l_priority)
    {
	/*
	** The priority of the drive being added is higher or equal to the
	** current first entry.  Thus we add it at the high end of the list
	** maintaining priority AND lru order.
	*/
	
	acpgeo.brnch$a_drives[drive].brnch$l_available = brnch_gl_busyDriveList[0] ;
	brnch_gl_busyDriveList[0] = drive ;
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_busyDriveList[0]) ;
	    brnch__logit("Exiting: addToBusy(%d)\n", drive) ;
	}
	return ;
    }
    
    /*
    ** Since we have already checked for addition at the beginning and end of
    ** the list, we are guaranteed that an insertion must occur somewhere in the
    ** middle.
    */
    
    thisDrive = brnch_gl_busyDriveList[0] ;
    
    for (;;)
    {
	lastDrive = thisDrive ;
	thisDrive = acpgeo.brnch$a_drives[thisDrive].brnch$l_available ;
	if (acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[drive].brnch$l_jbUnit].brnch$l_priority >=
	    acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[thisDrive].brnch$l_jbUnit].brnch$l_priority)
	{
	    /*
	    ** The priority of the jb unit in the new drive is greater or equal
	    ** to the priority of the jb unit in thisDrive.  Add drive before
	    ** thisDrive.
	    */
	    
	    acpgeo.brnch$a_drives[drive].brnch$l_available = thisDrive ;
	    acpgeo.brnch$a_drives[lastDrive].brnch$l_available = drive ;
	    if (brnch_isTraceArbitrationOn())
	    {
		printDriveList("    list", brnch_gl_busyDriveList[0]) ;
		brnch__logit("Exiting: addToBusy(%d)\n", drive) ;
	    }
	    return ;
	}
    }
} ;

static void
masterAnnounceGrantedAST (
    unsigned long drive)
{
    long				jbUnit ;
    unsigned long			status ;
    
    /*
    ** Tell the slaves WHICH cartridge is in the drive.
    */

    announceLSB[drive].brnch$l_drive = drive ;
    jbUnit = announceLSB[drive].brnch$l_jbUnit = acpgeo.brnch$a_drives[drive].brnch$l_jbUnit ;
    
    status = sys$enqw (0, LCK$K_CRMODE, &announceLSB[drive],
		       (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_VALBLK+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterAnnounceGrantedAST (unsigned long drive):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    status = brnch_dispatchRoutine(masterSetDwell, jbUnit) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterAnnounceGrantedAST (unsigned long drive):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    return ;
} ;

static void
masterDiskArbitration (
    brnch$r_sndcmd_message* message)
{
    switch(message -> brnch$l_function)
    {
	case BRNCH$K_SNDCMD_MESSAGE_STARTIO :
	    masterStartIO(message) ;
	    break ;

	case BRNCH$K_SNDCMD_MESSAGE_IOPOST :
	    masterIOPost(message) ;
	    break ;

	default:
	    {
		$DESCRIPTOR(xxx, "static void masterDiskArbitration (brnch$r_sndcmd_message* message):0") ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }
    }
} ;

static void
masterDwellExpired (
    unsigned long jbUnit)
{
    unsigned long			drive ;
    unsigned long			status ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: masterDwellExpired(%d)\n", jbUnit) ;
    }

    drive = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive ;

    /*
    ** Technically, it is possible to get multiple entries to this routine for a
    ** single jbUnit.  For example:
    **
    **	1. The jbUnit is dismounted.
    **	2. A startio for some newJBUnit is queued.
    **	3. The dwell for jbUnit expires.
    **	4. The pending list is scanned.
    **
    ** Which will result in calling removeFromBusy twice for the SAME jb unit.
    ** The test of removeFromBusy keeps us from blowing the available list up.
    **
    ** It is also possible to get a notification for a JB unit that is no longer
    ** in a drive, e.g., in the event of multiple start ios processed for a
    ** single jb unit under these circumstances we will get multiple timer ast's
    ** queued and [potentially] multiple entries to this routine which COULD
    ** result in the jb unit's removal from a drive BEFORE the processing of the
    ** NEXT timer AST.
    */
    
    if (drive == -1)
    {
        unsigned long                   i ;
        
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("JB unit %d is no longer in a drive\n", jbUnit) ;
	}
	
	/*
	** At this point, it's possible to have an orphan drive on the busy list.
	** This tends to happen mostly during EXPORT processing, but could (at
	** least in theory) happen in any number of other places.  Therefore the
	** fix is being put in here.
	*/
	
	for (i=0; i < acpgeo.brnch$l_numDrives; i++)
	{
	    if ((acpgeo.brnch$a_drives[i].brnch$r_flags.brnch$v_onBusy) &&
	        (acpgeo.brnch$a_drives[i].brnch$l_jbUnit == -1))
	    {
	        if(status = removeFromBusy(i))
	        {
		    /*
		    ** Since the drive was, indeed, on the busy list, lets also
		    ** assume that the inventory table has gotten skewed and
		    ** should be updated as well.
		    */
		    
		    struct brnch$r_inventory_table diskRecord ;

	            brnch_addToAvailable(i) ;		/* Make sure the orphaned drive goes back on the available list.	*/

		    status = brnch__readIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
		    if ($VMS_STATUS_SUCCESS(status))
		    {
			diskRecord.brnch$r_overlay.brnch$l_contents[i] = -1 ;
			brnch__updateIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
		    }

	            if (brnch_isTraceArbitrationOn())
	            {
	                brnch__logit("%08x = removeFromBusy(%d)\n", status, i) ;
	                brnch__logit("Drive %d had been orphaned\n", i) ;
	            }
	        }
	    }
	}
    }
    else
    {
	if (status = removeFromBusy(drive))
	{
	    brnch_addToAvailable(drive) ;
	    requeueJBUnits(acpgeo.brnch$a_drives[drive].brnch$l_jbUnit) ;
	}
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%08x = removeFromBusy(%d)\n", status, drive) ;
	}
    }

    status = brnch_dispatchRoutine(masterProcessPendingList,-1) ;
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_dispatchRoutine(masterProcessPendingList, -1)\n", status) ;
	brnch__logit("Exiting: masterDwellExpired(%d, %d)\n", jbUnit, drive) ;
    }
    return ;
} ;

static void
masterDwellExpiredAST (
    brnch$r_jbUnit* jbUnitPtr)
{
    unsigned long			jbUnit ;
    unsigned long			status ;
    
    jbUnit = jbUnitPtr -> brnch$l_index ;
    status = brnch_dispatchRoutine(masterDwellExpired, jbUnit) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterDwellExpiredAST(unsigned long jbUnit):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    return ;
} ;

static void
masterIOPost(
    brnch$r_sndcmd_message* message)
{
    unsigned long			jbUnit ;
    unsigned long			status ;
    
    jbUnit = message -> brnch$l_unit_number ;

    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_quiesceRequested = 1 ;

    status = sys$enq (0, LCK$K_EXMODE, &jbUnitQuiescedLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
		      0, masterSlaveIOQuiescedGrantedAST, jbUnit, 0, 0, 0) ;
							/* Arm notification from slaves of successful quiescence of the request	*/
							/* JB unit.								*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterIOPost(brnch$r_sndcmd_message* message):1") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    status = sys$enq (0, LCK$K_EXMODE, &quiesceJBUnitLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
		      0, masterQuiesceSlaveIOGrantedAST, jbUnit, 0, 0, 0) ;
							/* Tell the slaves to quiesce I/O on this JB unit.			*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterIOPost(brnch$r_sndcmd_message* message):2") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
} ;

static void
masterPeriodicIOStatusBlockingAST (
    unsigned long unused)
{

    /*
    ** FIX THIS: add this function.
    */
    
    return ;
}

/*
** A new entry in the pending list has been made.  This routine takes a look at
** the drive available list and sees if there is any way to get work done.
**
**	if the pending list is empty
**	then
**	    return
**	end if
**	
**	If no drive is available
**	then
**	    If the front of the pending list is of higher priority than a jb
**	    unit currently in a drive
**	    then
**		abort the dwell time for that drive and note the fact that the
**		dwell time has been aborted.
**	    end if
**	    return
**	end if
**
**	If the drive is full
**	then
**	    shutdown drive
**	    if the shutdown did not occur inline
**	    then
**		return
**	    end if
**	fake IOPOST
**	return
*/

static void
masterProcessPendingList(
    unsigned long unused)
{
    int					drive ;
    int					jbUnit ;
    unsigned long			status ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: masterProcessPendingList()\n") ;
    }
    
    if (brnch_gl_pendingJBUnits[0] == -1)
    {
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("Exiting: masterProcessPendingList() - no pending I/O\n") ;
	}
	return ;
    } ;

    if (ioPostInProgress)
    {
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("Exiting: masterProcessPendingList() - IO post in progress\n") ;
	}
	return ;					/* Ignore Scheduling attempts if there is an IOPOST in progress		*/
    }

    if (acpgeo.brnch$l_numSides != 1)
    {
	int				otherJBUnit ;
	unsigned long			unloadStatus ;

	jbUnit = brnch_gl_pendingJBUnits[0] ;		/* The only JB unit of interest is the first one.			*/

	/*
	** FIX THIS: should assume nsides > 2.
	*/
	
	otherJBUnit = jbUnit ^ 1 ;			/* Figure out the other JB unit number.					*/
	
	if ((drive = acpgeo.brnch$a_jbUnits[otherJBUnit].brnch$l_drive) != -1)
	{
	    /*
	    ** The requested JB unit's OTHER SIDE is already in a drive.  Figure
	    ** out if the other side's drive is still busy (or should be).
	    */
	    
	    if (acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onBusy)
	    {
		unloadStatus = guflg(otherJBUnit) ;
		if (brnch_isTraceArbitrationOn())
		{
		    brnch__logit("%08x = guflg(%d)\n", unloadStatus, otherJBUnit) ;
		}

		if ((acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_priority > acpgeo.brnch$a_jbUnits[otherJBUnit].brnch$l_priority) ||
		    (unloadStatus))
		{
		    /*
		    ** The requested JB unit priority is higher or the JB unit
		    ** currently in the drive has been unloaded.  Cancel the JB
		    ** unit currently in the drive and reschedule another pass
		    ** for later.
		    */
		    
		    status = sys$cantim (&acpgeo.brnch$a_jbUnits[otherJBUnit], 0) ;
							/* The dwell time can be cancelled for this drive since the priority is	*/
							/* lower than something in the pending list.				*/
		    if (brnch_isTraceArbitrationOn())
		    {
			brnch__logit("%08x = sys$cantim(%d, 0)\n", status, otherJBUnit) ;
		    }
		    sys$dclast (
			masterDwellExpiredAST, 
			&acpgeo.brnch$a_jbUnits[otherJBUnit], 
			0) ;

		    if (brnch_isTraceArbitrationOn())
		    {
			brnch__logit(
			    "Exiting: masterProcessPendingList() - drive %d contained %d which was lower priority than %d\n",
			    drive, otherJBUnit, jbUnit) ;
		    }
		}
		else
		{
		    if (brnch_isTraceArbitrationOn())
		    {
			brnch__logit(
			    "Exiting: masterProcessPendingList() - drive %d contained %d which was not lower priority than %d\n",
			    drive, otherJBUnit, jbUnit) ;
		    }
		}

		brnch_removeFromPending(jbUnit) ;	/* Can't process this puppy right now.					*/
		acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_pendingUponAvailable = 1 ;
							/* But when the side currently in the drive is done, queue it.		*/
		return ;
	    }

	    if (acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onAvailable)
	    {
		/*
		** The drive was on the available list (I'm checking to maintain
		** sanity, it really DOESN'T need to be checked).  That means
		** the JB unit in the drive must be quiesced.
		*/
		
		brnch_queueIOPost(otherJBUnit) ;

		if (brnch_isTraceArbitrationOn())
		{
		    brnch__logit("Exiting: masterProcessPendingList()\n") ;
		}

		return ;
	    }
	}
    }
    
    if ((drive = selectLRUAvailableDrive()) == -1)
    {
	unsigned long			freedADrive = 0 ;

	/*
	** We only care about the unload flag when all physical drives are busy.
	** Then we take a look and see if there has been an unload or available
	** for any drive and cancel all dwell times at that point.
	*/
	
	for (drive = 0;  drive < acpgeo.brnch$l_numDrives;  drive++)
	{
	    if (acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onDisabled)
	    {
		continue ;
	    }

	    if (status = guflg(acpgeo.brnch$a_drives[drive].brnch$l_jbUnit))
	    {
		brnch__cancelMasterDwell(drive) ;
		freedADrive++ ;
	    }

	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("%08x = guflg(%d)\n", status, acpgeo.brnch$a_drives[drive].brnch$l_jbUnit) ;
	    }
	}

	if (freedADrive)
	{
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit (
		    "Exiting: masterProcessPendingList() - no available drives but %d were unloaded or available\n",
		    freedADrive) ;
	    }
	    return ;
	}

	/*
	** In case nothing was freed up by the check for unload/available, check
	** to see if there is a lower priority jb unit in the drive on the
	** back of the busy drive list.  This will replace the LOWEST priority drive.
	*/
	
	if (acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[brnch_gl_busyDriveList[1]].brnch$l_jbUnit].brnch$l_priority <
	    acpgeo.brnch$a_jbUnits[brnch_gl_pendingJBUnits[0]].brnch$l_priority)
	{
	    brnch__cancelMasterDwell(brnch_gl_busyDriveList[1]) ;

	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit(
		    "Exiting: masterProcessPendingList() - no available drives but %d was lower priority than %d\n",
		    acpgeo.brnch$a_drives[brnch_gl_busyDriveList[1]].brnch$l_jbUnit, brnch_gl_pendingJBUnits[0]) ;
	    }
	}
	else
	{
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("Exiting: masterProcessPendingList() - no available drives\n") ;
	    }
	}
	return ;					/* A return is all that's required since the expiration of the dwell	*/
							/* time will put the drive onto the available list and trigger another	*/
							/* scan of the pending list.						*/
    }

    brnch_queueIOPost(acpgeo.brnch$a_drives[drive].brnch$l_jbUnit) ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: masterProcessPendingList()\n") ;
    }

    return ;
}

/*
** Called when all the slaves have seen the request to quiesce I/O for a
** specific JB unit.  All it does is return the lock to CR mode.
*/

static void
masterQuiesceSlaveIOGrantedAST(
    unsigned long jbUnit)
{
    unsigned long			status ;
    
    status = sys$enqw (0, LCK$K_CRMODE, &quiesceJBUnitLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterQuiesceSlaveIOGrantedAST(unsigned long jbUnit):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    return ;
} ;

static void
masterSetDwell (
    unsigned long jbUnit)
{
    long				dwell ;
    unsigned long			status ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: masterSetDwell(%d)\n", jbUnit) ;
    }

    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_announceInProgress = 0 ;
							/* Note that announcing a jbUnit in a drive is over.			*/
    dwell = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_dwell ;
    if (dwell == BRNCH_K_DEFAULT_UNIT_DWELL)
    {
	dwell = brnch_gl_dwell ;
    }

    if (dwell)
    {
	unsigned long			opCode  = LIB$K_DELTA_SECONDS ;
	long				delta[2] ;
	long				time[2] ;
	
	lib$cvt_to_internal_time (&opCode, &dwell, &delta) ;
	sys$gettim (&time) ;
	lib$add_times (&time, &delta, &time) ;
	status = sys$setimr (0, &time, masterDwellExpiredAST, &acpgeo.brnch$a_jbUnits[jbUnit], 0) ;
	if (brnch_isTraceArbitrationOn())
	{
	    char			buffer[23] ;
	    struct dsc$descriptor_s	bufferDesc = {sizeof(buffer), DSC$K_DTYPE_T, DSC$K_CLASS_S, buffer} ;

	    sys$asctim (&bufferDesc.dsc$w_length, &bufferDesc, &time, 0) ;
	    brnch__logit(
		"%08x = sys$setimr (0, %0.*s, masterDwellExpiredAST, %d, 0)\n",
		status, bufferDesc.dsc$w_length, bufferDesc.dsc$a_pointer, jbUnit) ;
	}
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    $DESCRIPTOR(xxx,"static void masterSetDwell(unsigned long jbUnit):0") ;
	    lib$signal(status, 0) ;
	    lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	}
    }
    else
    {
	status = sys$dclast (masterDwellExpiredAST, &acpgeo.brnch$a_jbUnits[jbUnit], 0) ;
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%08x = sys$dclast (masterDwellExpiredAST, %d, 0)\n", status, jbUnit) ;
	}
    }

    status = brnch_dispatchRoutine(masterProcessPendingList, -1) ;
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_dispatchRoutine(masterProcessPendingList, -1)\n", status) ;
	brnch__logit("exiting: masterSetDwell(%d, %d)\n", jbUnit, dwell) ;
    }
    return ;
} ;

/*
** At this point, the specified JB unit has been quiesced by all slaves.  If it
** is in a drive, it SHOULD be removed, but we'll want to check to see if the
** front of the pending list is on the "other" side.  If so, we just do a flip.
**
** If the posted jb unit is not in a drive
** then
**	move the jb unit to the first drive in the available list.
** end if
*/

static void
masterSlaveIOQuiesced(
    unsigned long jbUnit)
{
    struct brnch$r_inventory_table	diskRecord ;
    int					drive ;
    int					flip ;
    int					newJBUnit ;
    unsigned long			status ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
    }

    ioPostInProgress = 0 ;				/* At long last, the I/O POST operation is finally complete.		*/
    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_quiesceRequested = 0 ;

    newJBUnit = brnch_gl_pendingJBUnits[0] ;		/* Always take the first entry in the pending list.			*/
    flip = 0 ;						/* Don't flip the cartridge in the "drive".				*/

    if ((drive = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive) != -1)
    {
	if (jbUnit == newJBUnit)
	{
	    /*
	    ** The system selected a drive to quiesce and between that time and
	    ** "now" a STARTIO has been received that queued itself to the front
	    ** of the pending list for this drive.  Since we are guaranteed that
	    ** this is the highest priority request (the pending list is stored
	    ** in priority order) all we need to do is reconnect the drive.  No
	    ** robot motion is necessary.
	    */
	    
	    brnch_removeFromAvailable(drive) ;		/* Remove the drive from the available list.				*/
	    if (!brnch_removeFromPending(newJBUnit))	/* Remove the new jb unit from the pending list.			*/
	    {
		$DESCRIPTOR(xxx,"static void static void masterSlaveIOQuiesced(unsigned long jbUnit):0") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }
	    addToBusy(drive) ;				/* Add the drive to the busy list.					*/

	    status = turnOnPhysicalDrive(drive, newJBUnit) ;
							/* Turn on the underlying physical device.				*/
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("%08x = turnOnPhysicalDrive(%d, %d)\n", status, drive, newJBUnit) ;
	    }
	    
	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		turnOnPhysicalDriveFailed(drive, newJBUnit) ;

		if (brnch_isTraceArbitrationOn())
		{
		    brnch__logit("Exiting: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
		}

		return ;
	    }
	    
	    /*
	    ** Let the slaves know that there is a cartridge in the drive.
	    */

	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_announceInProgress = 1 ;
							/* Note that announcing a jbUnit in a drive is in progress.		*/
	    status = sys$enq (0, LCK$K_EXMODE, &announceLSB[drive],
			      (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_VALBLK+LCK$M_NODLCKWT), 0,
			      0, masterAnnounceGrantedAST, drive, 0, 0, 0) ;
	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		$DESCRIPTOR(xxx,"static void static void masterSlaveIOQuiesced(unsigned long jbUnit):1") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }
	    
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("Exiting: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
	    }
	    return ;
	}
	
	/*
	** This ISN'T a fake IOPOST for a JB unit that isn't in a drive, so we
	** should get the drive unloaded.
	*/
	
	brnch_unloadPhysicalDevice(drive) ;

	flip = brnch_isFlipNecessary(jbUnit, newJBUnit) ;
							/* Set the flag indicating that the drive should be flipped.		*/
	if (!flip)
	{
	    if (acpgeo.brnch$l_pioneer604x == 0)
	    {
		status = brnch_driveToVolume(drive, jbUnit) ;
							/* The jb unit was actually IN a drive, just put it away.		*/
	    }
	    else
	    {
		status = SS$_NORMAL ;
	    }

	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		brnch__logit(
		    "Failed to unload volume %d (slot %d) from drive %d (%0.*s)",
		    jbUnit,
		    jbUnit/acpgeo.brnch$l_numSides,
		    drive,
		    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$w_length,
		    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$a_pointer) ;
		lib$signal (status, 0) ;
	    }
	}

	acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive = -1 ;
	acpgeo.brnch$a_drives[drive].brnch$l_jbUnit = -1 ;
							/* note that the drive no longer contains the JB unit and that the JB	*/
							/* unit is no longer in a drive.					*/
    }
    else
    {
	drive = selectLRUAvailableDrive() ;		/* The drive to contain the NEW jb unit is the least recently used	*/
							/* physical drive  .							*/
    }

    if (flip)
    {
	status = brnch_flipInDrive(drive) ;		/* Move the jb unit to the requested drive.				*/
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%08x = brnch_flipInDrive(%d)\n", status, drive) ;
	}
    }
    else
    {
	if (acpgeo.brnch$l_pioneer604x == 0)
	{
	    status = brnch_volumeToDrive(newJBUnit, drive) ;/* Move the jb unit to the requested drive.				*/
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("%08x = brnch_volumeToDrive(%d, %d)\n", status, newJBUnit, drive) ;
	    }
	}
	else
	{
	    status = SS$_NORMAL ;
	}
    }

    if (!$VMS_STATUS_SUCCESS(status))
    {
	brnch__logit(
	    (flip ? "Failed to flip volume %d (slot %d) in drive %d (%0.*s)"
		  : "Failed to load volume %d (slot %d) into drive %d (%0.*s)"),
	    newJBUnit,
	    newJBUnit/acpgeo.brnch$l_numSides,
	    drive,
	    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$w_length,
	    acpgeo.brnch$a_drives[drive].brnch$a_name -> dsc$a_pointer) ;
	lib$signal (status, 0) ;
    }

    /*
    ** Update the inventory table.
    */
    
    status = brnch__readIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
    if ($VMS_STATUS_SUCCESS(status))
    {
	diskRecord.brnch$r_overlay.brnch$l_contents[drive] = newJBUnit ;
	brnch__updateIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
    }

    brnch_removeFromAvailable(drive) ;			/* Remove the drive from the available list.				*/
    if (!brnch_removeFromPending(newJBUnit))		/* Remove the new jb unit from the pending list.			*/
    {
	$DESCRIPTOR(xxx,"static void static void masterSlaveIOQuiesced(unsigned long jbUnit):2") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    acpgeo.brnch$a_drives[drive].brnch$l_jbUnit = newJBUnit ;
    acpgeo.brnch$a_jbUnits[newJBUnit].brnch$l_drive = drive ;
							/* Link the drive and the JB unit together. This MUST be done before	*/
							/* adding the drive to the busy list.					*/
    addToBusy(drive) ;					/* Add the drive to the busy list.					*/

    status = turnOnPhysicalDrive(drive, newJBUnit) ;	/* Turn on the underlying physical device.				*/
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = turnOnPhysicalDrive(%d, %d)\n", status, drive, newJBUnit) ;
    }
    
    if (!$VMS_STATUS_SUCCESS(status))
    {
	turnOnPhysicalDriveFailed(drive, newJBUnit) ;

	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("Exiting: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
	}

	return ;
    }
    
    /*
    ** Let the slaves know that there is a cartridge in the drive.
    */

    acpgeo.brnch$a_jbUnits[newJBUnit].brnch$r_flags.brnch$v_announceInProgress = 1 ;
							/* Note that announcing a jbUnit in a drive is in progress.		*/
    status = sys$enq (0, LCK$K_EXMODE, &announceLSB[drive],
		      (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_VALBLK+LCK$M_NODLCKWT), 0,
		      0, masterAnnounceGrantedAST, drive, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void masterSlaveIOQuiesced(unsigned long jbUnit):3") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
    }
    return ;
} ;


static void
masterSlaveIOQuiescedGrantedAST(
    unsigned long jbUnit)
{
    unsigned long			status ;
    
    status = sys$enqw (0, LCK$K_CRMODE, &jbUnitQuiescedLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterSlaveIOQuiescedGrantedAST(unsigned long jbUnit):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    status = brnch_dispatchRoutine(masterSlaveIOQuiesced, jbUnit) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void masterSlaveIOQuiescedGrantedAST(unsigned long jbUnit):1") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    
    return ;
} ;

static void
masterStartIO(
    brnch$r_sndcmd_message* message)
{
    unsigned long			drive ;
    unsigned long			jbUnit ;
    unsigned long			status ;
    
    jbUnit = message -> brnch$l_unit_number ;
    if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onExported)
    {
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_startIOSeen = 1 ;
	
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("JB unit %d has been exported.  The startIO has been logged.\n", jbUnit) ;
	}
    }
    else if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_onPending)
    {
	/*
	** In the case of clusters, it is quite possible for a single JB unit
	** to see multiple start io requests.  Make sure the JB unit isn't
	** already on the pending list BEFORE putting it there.
	*/
	
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("JB unit %d is already on the pending list\n", jbUnit) ;
	}
    }
    else if ((drive = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive) != -1)
    {
	/*
	** It is also possible to get a startio for a JB unit that is already
	** in a drive.  We get this under normal operation, but we ALSO get
	** this when we are doing failover, so we cannot simply ignore the
	** startio.  Basically, all we do is to announce the jb unit in the
	** drive once more.
	*/

	if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_quiesceRequested)
	{
	    /*
	    ** This unit is actively being quiesced.  Just put it on the
	    ** pending queue and let the quiesce code check to see if anything
	    ** special needs to be done.
	    */
	    
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("JB unit %d is being quiesced, add to pending\n", jbUnit) ;
	    }

	    brnch_addToPending(jbUnit) ;
	}
	else if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_announceInProgress)
	{
	    /*
	    ** if the jb unit is not bneing quiesced, it MAY be in the
	    ** process of being announced, in which case, we may safely
	    ** ignore the startio.
	    */

	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("JB unit %d is being announced, ignore startio\n", jbUnit) ;
	    }
	}
	else
	{
	    status = turnOnPhysicalDrive(drive, jbUnit) ;
							/* Turn on the underlying physical device.				*/
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("%08x = turnOnPhysicalDrive(%d, %d)\n", status, drive, jbUnit) ;
	    }
	    
	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		turnOnPhysicalDriveFailed(drive, jbUnit) ;

		if (brnch_isTraceArbitrationOn())
		{
		    brnch__logit("Exiting: masterSlaveIOQuiesced(%d)\n", jbUnit) ;
		}

		return ;
	    }
	    
	    /*
	    ** Let the slaves know that there is a cartridge in the drive.
	    */

	    acpgeo.brnch$a_jbUnits[jbUnit].brnch$r_flags.brnch$v_announceInProgress = 1 ;
						    /* Note that announcing a jbUnit in a drive is in progress.		*/
	    status = sys$enq (0, LCK$K_EXMODE, &announceLSB[drive],
			      (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_VALBLK+LCK$M_NODLCKWT), 0,
			      0, masterAnnounceGrantedAST, drive, 0, 0, 0) ;
	    if (!$VMS_STATUS_SUCCESS(status))
	    {
		$DESCRIPTOR(xxx,"static void masterStartIO(brnch$r_sndcmd_message* message):0") ;
		lib$signal(status, 0) ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }
	}
    }
    else
    {
	brnch_addToPending(jbUnit) ;
    }
    status = brnch_dispatchRoutine(masterProcessPendingList,-1) ;
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_dispatchRoutine(masterProcessPendingList, -1)\n", status) ;
    }
    return ;
} ;

static void
masterStartIOBlockingAST (
    unsigned long jbUnit)
{
    unsigned long			status ;

    status = sys$enqw (0, LCK$K_NLMODE, &startIOLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_VALBLK), 0, 
		       0, 0, 0, 0, 0, 0) ;		/* Let the slave continue.						*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(location, "static void masterStartIOBlockingAST (unsigned long jbUnit):0") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    status = sys$enq (0, LCK$K_CRMODE, &startIOLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_VALBLK+LCK$M_NODLCKWT), 0, 
		      0, 0, jbUnit, masterStartIOBlockingAST, 0, 0) ;
							/* Let the slave continue.						*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(location, "static void masterStartIOBlockingAST (unsigned long jbUnit):1") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    status = brnch_sendCommandMessage(BRNCH$K_SNDCMD_MESSAGE_STARTIO, 0, jbUnit, 0, 0, gatim()) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(location, "static void masterStartIOBlockingAST (unsigned long jbUnit):2") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    return ;
}

static void
printDriveList(
    char* listName,
    long entry)
{
    unsigned long			i ;
    
    brnch__logit("%s = ", listName) ;

    for (i = 0; entry != -1; i++)
    {
	char				e[256] ;
	unsigned long			index ;

	if (i)
	{
	    index = 2 ;
	    strcpy (e, ", ") ;
	}
	else
	{
	    index = 0 ;
	}

	index += sprintf(&e[index], "%d(%d", entry, acpgeo.brnch$a_drives[entry].brnch$l_jbUnit) ;
	if (acpgeo.brnch$a_drives[entry].brnch$l_jbUnit == -1)
	{
	    strcat (e, ")") ;
	}
	else
	{
	    sprintf(&e[index], ", %d, %d)",
		acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[entry].brnch$l_jbUnit].brnch$l_dwell,
		acpgeo.brnch$a_jbUnits[acpgeo.brnch$a_drives[entry].brnch$l_jbUnit].brnch$l_priority) ;
	}

	brnch__logit("%s", e) ;

	entry = acpgeo.brnch$a_drives[entry].brnch$l_available ;
    }

    if (i)
    {
	brnch__logit("\n") ;
    }
    else
    {
	brnch__logit("empty\n") ;
    }
}

static void
printJBUnitList(
    char* listName,
    long entry)
{
    unsigned long			i ;
    
    brnch__logit("%s = ", listName) ;

    for (i = 0; entry != -1; i++)
    {
	char				e[256] ;
	unsigned long			index ;
	
	if (i)
	{
	    index = 2 ;
	    strcpy(e, ", ") ;
	}
	else
	{
	    index = 0 ;
	}
	
	sprintf(&e[index], "%d(%d, %d, %d)",
	    entry,
	    acpgeo.brnch$a_jbUnits[entry].brnch$l_drive,
	    acpgeo.brnch$a_jbUnits[entry].brnch$l_dwell,
	    acpgeo.brnch$a_jbUnits[entry].brnch$l_priority) ;

	brnch__logit("%s", e) ;

	entry = acpgeo.brnch$a_jbUnits[entry].brnch$l_pending ;
    }

    if (i)
    {
	brnch__logit("\n") ;
    }
    else
    {
	brnch__logit("empty\n") ;
    }
}

unsigned long
brnch_queueIOPost(
    unsigned long jbUnit)
{
    /*
    ** Called to force an IO post for a specific JB unit OR for the head of the
    ** pending list.
    */
    
    unsigned long			status ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("unsigned long brnch_queueIOPost(%d)\n", jbUnit) ;
    }
    
    ioPostInProgress = 1 ;				/* Note that there is an I/O post outstanding.  This, in effect, turns	*/
							/* off arbitration until an I/O post is processed by the arbitration	*/
							/* system.  This is a simplifying case.  I believe that I could get	*/
							/* many things in process at once, but at a substantial increase in	*/
							/* complexity.								*/
    if (jbUnit != -1)
    {
	/*
	** Otherwise, quiesce the unit.
	*/
	
	status = ibsyck(jbUnit) ;
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%08x = ibsyck(%d)\n", status, jbUnit) ;
	}
	if (status == 2)
	{
	    /*
	    ** The unit had I/O outstanding, so the driver will, eventually,
	    ** issue an I/O POST command which we will process.
	    */
	    
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("1 = brnch_queueIOPost(%d)\n", jbUnit) ;
	    }
	    return 1 ;					/* The driver will be sending an IOPOST so we don't have to fake one.	*/
	}
    }
    else
    {
	jbUnit = brnch_gl_pendingJBUnits[0] ;		/* There wasn't anything in the drive of "choice", so we just fake an	*/
							/* I/O post for the JB unit that we want to put into the drive.		*/
							
    } ;

    if (jbUnit != -1)
    {
	/*
	** Since this can be called when the pending JBUnits list is empty, this
	** keeps us from buggering up the rest of the disk arbitration.
	*/
	
	status = brnch_sendCommandMessage(BRNCH$K_SNDCMD_MESSAGE_IOPOST, 0, jbUnit, 0, 0 , gatim()) ;
							/* This sends a fake I/O POST message that will eventually get		*/
							/* processed by the disk arbitration AS IF the unit had been busy all	*/
							/* along.								*/
	if (!$VMS_STATUS_SUCCESS(status))
	{
	    $DESCRIPTOR(xxx,"unsigned long brnch_queueIOPost():0") ;
	    lib$signal(status, 0) ;
	    lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	}

	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("Exiting: 2 = brnch_queueIOPost(%d)\n", jbUnit) ;
	}

	return 2 ;
    }
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: 0 = brnch_queueIOPost(%d)\n", jbUnit) ;
    }

    ioPostInProgress = 0 ;				/* I'm pretty sure this can never happen but just in case, if an I/O	*/
							/* post doesn't get queued, then make sure that no notification of an	*/
							/* I/O post is recorded in the long term history of the disk		*/
							/* arbitration module.							*/
    return 0 ;
}

void
brnch_removeFromAvailable(
    unsigned long drive)
{
    int					lastDrive ;
    int					thisDrive ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: brnch_removeFromAvailable(%d)\n", drive) ;
	printDriveList("    list", brnch_gl_availableDriveList[0]) ;
    }
    
    acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onAvailable = 0 ;
    
    if (brnch_gl_availableDriveList[0] == -1)
    {
	$DESCRIPTOR(xxx,"void brnch_removeFromAvailable(unsigned long drive):0") ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    if (brnch_gl_availableDriveList[0] == drive)
    {
	if (brnch_gl_availableDriveList[1] == drive)
	{
	    /*
	    ** This is the only element on the list, so reset the list to
	    ** indicate that it is empty.  For drill, make sure the available
	    ** pointer is cleared as well.
	    */
	    
	    brnch_gl_availableDriveList[0] = brnch_gl_availableDriveList[1] = acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	}
	else
	{
	    /*
	    ** This is the first element on the list, so remove it from the
	    ** list.  For drill, make sure the available pointer is cleared as
	    ** well.
	    */
	    
	    brnch_gl_availableDriveList[0] = acpgeo.brnch$a_drives[drive].brnch$l_available ;
	    acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	}
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_availableDriveList[0]) ;
	    brnch__logit("Exiting: brnch_removeFromAvailable(%d)\n", drive) ;
	}
	return ;
    } ;

    /*
    ** The requested drive isn't the first one on the list, so we have to look
    ** for it.
    */

    thisDrive = brnch_gl_availableDriveList[0] ;

    for (;;)
    {
	lastDrive = thisDrive ;
	thisDrive = acpgeo.brnch$a_drives[thisDrive].brnch$l_available ;
	if (thisDrive == -1)
	{
	    $DESCRIPTOR(xxx,"void brnch_removeFromAvailable(unsigned long drive):1") ;
	    lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	}

	if (thisDrive == drive)
	{
	    acpgeo.brnch$a_drives[lastDrive].brnch$l_available = acpgeo.brnch$a_drives[thisDrive].brnch$l_available ;
	    if (acpgeo.brnch$a_drives[lastDrive].brnch$l_available == -1)
	    {
		brnch_gl_availableDriveList[1] = lastDrive ;
	    } ;
	    acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	    if (brnch_isTraceArbitrationOn())
	    {
		printDriveList("    list", brnch_gl_availableDriveList[0]) ;
		brnch__logit("Exiting: brnch_removeFromAvailable(%d)\n", drive) ;
	    }
	    return ;
	}
    }
} ;

static unsigned long
removeFromBusy (
    unsigned long drive)
{
    int					lastDrive ;
    int					thisDrive ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: removeFromBusy(%d)\n", drive) ;
	printDriveList("    list", brnch_gl_busyDriveList[0]) ;
    }
    
    acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onBusy = 0 ;

    if (brnch_gl_busyDriveList[0] == -1)
    {	
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_busyDriveList[0]) ;
	    brnch__logit("Exiting: removeFromBusy(%d)\n", drive) ;
	}
	return 0 ;					/* Nothing has been removed from this list.				*/
    }

    if (brnch_gl_busyDriveList[0] == drive)
    {
	if (brnch_gl_busyDriveList[1] == drive)
	{
	    /*
	    ** This is the only element on the list, so reset the list to
	    ** indicate that it is empty.  For drill, make sure the available
	    ** pointer is cleared as well.
	    */
	    
	    brnch_gl_busyDriveList[0] = brnch_gl_busyDriveList[1] = acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	}
	else
	{
	    /*
	    ** This is the first element on the list, so remove it from the
	    ** list.  For drill, make sure the available pointer is cleared as
	    ** well.
	    */
	    
	    brnch_gl_busyDriveList[0] = acpgeo.brnch$a_drives[drive].brnch$l_available ;
	    acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	}
	if (brnch_isTraceArbitrationOn())
	{
	    printDriveList("    list", brnch_gl_busyDriveList[0]) ;
	    brnch__logit("Exiting: removeFromBusy(%d)\n", drive) ;
	}
	return 1 ;					/* Something was removed from this list.				*/
    } ;

    /*
    ** The requested drive isn't the first one on the list, so we have to look
    ** for it.
    */

    thisDrive = brnch_gl_busyDriveList[0] ;

    for (;;)
    {
	lastDrive = thisDrive ;
	thisDrive = acpgeo.brnch$a_drives[thisDrive].brnch$l_available ;
	if (thisDrive == -1)
	{
	    if (brnch_isTraceArbitrationOn())
	    {
		printDriveList("    list", brnch_gl_busyDriveList[0]) ;
		brnch__logit("Exiting: removeFromBusy(%d)\n", drive) ;
	    }
	    return 0 ;					/* Nothing was removed from this list.					*/
	}

	if (thisDrive == drive)
	{
	    acpgeo.brnch$a_drives[lastDrive].brnch$l_available = acpgeo.brnch$a_drives[thisDrive].brnch$l_available ;
	    if (acpgeo.brnch$a_drives[lastDrive].brnch$l_available == -1)
	    {
		brnch_gl_busyDriveList[1] = lastDrive ;
	    } ;
	    acpgeo.brnch$a_drives[drive].brnch$l_available = -1 ;
	    if (brnch_isTraceArbitrationOn())
	    {
		printDriveList("    list", brnch_gl_busyDriveList[0]) ;
		brnch__logit("Exiting: removeFromBusy(%d)\n", drive) ;
	    }
	    return 1 ;					/* Something was removed from this list.				*/
	}
    }
} ;

static void
requeueJBUnits (
    unsigned long jbUnit)
{
    int					i ;
    unsigned long			status ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entered: requeueJBUnits(%d)\n", jbUnit) ;
    }

    /*
    ** FIX THIS: don't assume nsides = 1 or 2
    */

    if (acpgeo.brnch$l_numSides == 1)
    {
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("Exiting: requeueJBUnits(%d) - number of sides is 1\n", jbUnit) ;
	}
	return ;
    }

    jbUnit = jbUnit & -2 ;

    for (i = 0;  i < acpgeo.brnch$l_numSides;  i++)
    {
	if (acpgeo.brnch$a_jbUnits[jbUnit + i].brnch$r_flags.brnch$v_pendingUponAvailable)
	{
	    acpgeo.brnch$a_jbUnits[jbUnit + i].brnch$r_flags.brnch$v_pendingUponAvailable = 0 ;
	    brnch_addToPending(jbUnit + i) ;
	    status = brnch_dispatchRoutine(masterProcessPendingList,-1) ;
	    if (brnch_isTraceArbitrationOn())
	    {
		brnch__logit("%08x = brnch_dispatchRoutine(masterProcessPendingList, -1)\n", status) ;
		brnch__logit("Exiting: requeueJBUnits(%d) - unit requeued\n", jbUnit + i) ;
	    }
	    return ;
	}
    }

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Exiting: requeueJBUnits(%d) - no jbUnit requeued\n", jbUnit) ;
    }
    return ;
}

static int
selectLRUAvailableDrive()
{
    int					jbUnit ;
    int					selectedDrive = brnch_gl_availableDriveList[0];
    unsigned long			selectedDriveTime[2] ;
    int					thisDrive ;
    unsigned long			thisDriveTime[2] ;
    unsigned long			time[2] ;
    
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("static int selectLRUAvailableDrive()\n") ;
    }
    
    if (selectedDrive == -1)
    {
exit :
    
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%d = static int selectLRUAvailableDrive()\n", selectedDrive) ;
	}

	return selectedDrive ;
    }

    if (selectedDrive == brnch_gl_availableDriveList[1])
    {
	goto exit ;					/* There is only one drive on the list, so don't go any further.	*/
    }

    if ((jbUnit = acpgeo.brnch$a_drives[selectedDrive].brnch$l_jbUnit) == -1)
    {
	goto exit ;					/* The drive doesn't have a jbUnit in it at the moment, so just use it!	*/
    }

    /*
    ** There is more than one drive on the available list, scan the list
    ** looking for the drive whose last I/O time is furthest away in the past.
    ** The assumption we're making here is that a drive that hasn't done I/O
    ** "recently" won't do it "soon".
    */

    brnch_getSysIOTime(jbUnit, selectedDriveTime) ;
    
    for (thisDrive = acpgeo.brnch$a_drives[selectedDrive].brnch$l_available;
	 thisDrive != -1;
	 thisDrive = acpgeo.brnch$a_drives[thisDrive].brnch$l_available)
    {
	unsigned long			status ;

	if ((jbUnit = acpgeo.brnch$a_drives[thisDrive].brnch$l_jbUnit) == -1)
	{
	    selectedDrive = thisDrive ;			/* This drive doesn't have a jbUnit in it at the moment, so just use	*/
							/* it.									*/
	    goto exit ;
	}

	brnch_getSysIOTime(jbUnit, thisDriveTime) ;

	status = lib$sub_times (selectedDriveTime, thisDriveTime, time) ;
	if (status == LIB$_NORMAL)
	{
	    /*
	    ** When lib$sub_times returns normal, the times have the following
	    ** relationship:
	    **
	    **	selectedDriveTime > thisDriveTime
	    **
	    ** which means that thisDriveTime is further back in the past.
	    ** Select thisDrive as the candidate to be returned and keep going.
	    */

	    selectedDrive = thisDrive ;
	    memcpy (selectedDriveTime, thisDriveTime, sizeof(thisDriveTime)) ;
	}
    }

    goto exit ;						/* selectedDrive contains the physical drive index to be used at this	*/
							/* point.								*/
} ;

/*
** The master is announcing that a specific drive now contains a specific JB
** unit.
*/

static void
slaveAnnounce(
    brnch$r_announce_message* message)
{
    long				drive ;
    long				jbUnit ;
    long				newJBUnit ;
    unsigned long			status ;
    
    drive = message -> brnch$l_drive ;
    newJBUnit = message -> brnch$l_unit ;
    jbUnit = acpgeo.brnch$a_drives[drive].brnch$l_jbUnit ;

    if (jbUnit != -1)
    {
	acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive = -1 ;
	acpgeo.brnch$a_drives[drive].brnch$l_jbUnit = -1 ;
							/* note that the drive no longer contains the JB unit and that the JB	*/
							/* unit is no longer in a drive.					*/
    }        

    status = brnch_removeFromPending(newJBUnit) ;	/* Remove the new JB unit from the pending list.			*/
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_removeFromPending(%d)\n", status, newJBUnit) ;
    }
    acpgeo.brnch$a_drives[drive].brnch$l_jbUnit = newJBUnit ;
    acpgeo.brnch$a_jbUnits[newJBUnit].brnch$l_drive = drive ;
							/* We don't maintain the available or the busy list, but we still keep	*/
							/* track of what's in the drives just to get startio sane during	*/
							/* restart.								*/
    status = turnOnPhysicalDrive(drive, newJBUnit) ;	/* Turn on the underlying physical device.				*/
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = turnOnPhysicalDrive(%d, %d)\n", status, drive, newJBUnit) ;
    }
    return ;
} ;

static void
slaveAnnounceBlockingAST(
    unsigned long drive)
{
    unsigned long			status ;

    status = sys$enqw (0, LCK$K_NLMODE, &announceLSB[drive],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;		/* Let the master write the lock status block.				*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void slaveAnnounceBlockingAST(unsigned long drive):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    status = sys$enq (0, LCK$K_CRMODE, &announceLSB[drive],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
		      0, slaveAnnounceGrantedAST, drive, 0, 0, 0) ;
							/* Read the lock status block once the master has written it.		*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void slaveAnnounceBlockingAST(unsigned long drive):1") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
}

static void
slaveAnnounceGrantedAST(
    unsigned long drive)    
{
    unsigned long			status ;

    status = brnch_sendCommandMessage (BRNCH$K_SNDCMD_MESSAGE_ANNOUNCE, announceLSB[drive].brnch$l_drive, announceLSB[drive].brnch$l_jbUnit, 0, 0, gatim()) ;
							/* Announce the drive being loaded to the slave.			*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void slaveAnnounceGrantedAST(unsigned long drive):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    status = sys$enq (0, LCK$K_CRMODE, &announceLSB[drive],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		      0, 0, drive, slaveAnnounceBlockingAST, 0, 0) ;
							/* Reset the blocking AST for the next announcement for this drive.	*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void slaveAnnounceGrantedAST(unsigned long drive):1") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
}

static void
slaveDiskArbitration (
    brnch$r_sndcmd_message* message)
{
    
    switch(message -> brnch$l_function)
    {
	case BRNCH$K_SNDCMD_MESSAGE_STARTIO :
	    slaveStartIO(message) ;
	    break ;

	case BRNCH$K_SNDCMD_MESSAGE_QUIESCE :
	    slaveQuiesce(message) ;
	    break ;
	    
	case BRNCH$K_SNDCMD_MESSAGE_IOPOST :
	    slaveIOPost(message) ;
	    break ;

	case BRNCH$K_SNDCMD_MESSAGE_ANNOUNCE :
	    slaveAnnounce(message) ;
	    break ;
	    
	default:
	    {
		$DESCRIPTOR(xxx, "static void slaveDiskArbitration (brnch$r_sndcmd_message* message):0") ;
		lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
	    }
    }
} ;

static void
slaveIOPost(
    brnch$r_sndcmd_message* message)
{
    long				drive ;
    long				jbUnit ;
    unsigned long			status ;
    
    jbUnit = message -> brnch$l_unit_number ;
    if ((drive = acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive) != -1)
    {
	/*
	** This ISN'T a fake IOPOST for a JB unit that isn't in a drive, so we
	** should get the drive unloaded.
	*/
	
	brnch_unloadPhysicalDevice(drive) ;
    }

    status = sys$enqw (0, LCK$K_NLMODE, &jbUnitQuiescedLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;
							/* Notify master that the unit is now quiesced by giving up the lock.	*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void slaveIOPost(brnch$r_sndcmd_message* message):1") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    status = sys$enqw (0, LCK$K_CRMODE, &jbUnitQuiescedLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;		/* Get back into line for the NEXT notification.			*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void slaveIOPost(brnch$r_sndcmd_message* message):2") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
} ;

static void
slaveQuiesce(
    brnch$r_sndcmd_message* message)
{
    long				jbUnit ;
    unsigned long			status ;
    
    jbUnit = message -> brnch$l_unit_number ;
    if (acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive != -1)
    {
	status = ibsyck(jbUnit) ;
	if (brnch_isTraceArbitrationOn())
	{
	    brnch__logit("%08x = ibsyck(%d)\n", status, jbUnit) ;
	}
	if (status == 2)
	{
	    return ;					/* The driver will be sending an IOPOST so we don't have to fake one.	*/
	}
    }

    status = brnch_sendCommandMessage(BRNCH$K_SNDCMD_MESSAGE_IOPOST, 0, jbUnit, 0, 0 , gatim()) ;
							/* This sends a fake I/O POST message that will eventually get		*/
							/* processed by the disk arbitration AS IF the unit had been busy all	*/
							/* along.								*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void static void slaveQuiesce(brnch$r_sndcmd_message* message):0") ;
	lib$signal(status, 0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }
    return ;
} ;

static void
slaveQuiesceJBUnitBlockingAST(
    unsigned long jbUnit)
{
    unsigned long			status ;
    
    /*
    ** Let the master proceed.
    */
    
    status = sys$enqw (0, LCK$K_NLMODE, &quiesceJBUnitLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_VALBLK), 0, 
		       0, 0, 0, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location, "static void slaveQuiesceJBUnitBlockingAST(unsigned long jbUnit):0") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    status = sys$enq (0, LCK$K_CRMODE, &quiesceJBUnitLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_VALBLK), 0, 
		      0, 0, jbUnit, slaveQuiesceJBUnitBlockingAST, 0, 0) ;
							/* Re-arm the quiesce notification.					*/
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location, "static void slaveQuiesceJBUnitBlockingAST(unsigned long jbUnit):1") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }

    status = brnch_sendCommandMessage(BRNCH$K_SNDCMD_MESSAGE_QUIESCE, 0, jbUnit, 0, 0 , gatim()) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	static $DESCRIPTOR(location, "static void slaveQuiesceJBUnitBlockingAST(unsigned long jbUnit):2") ;
	lib$signal (status, 0) ;
	lib$signal (BRNCH$_INTLOGERR, 1, &location) ;
    }
} ;

static void
slaveStartIO(
    brnch$r_sndcmd_message* message)
{
    long				jbUnit ;
    unsigned long			status ;
    
    jbUnit = message -> brnch$l_unit_number ;
    brnch_addToPending(jbUnit) ;

    status = sys$enq (0, LCK$K_EXMODE, &startIOLSB[jbUnit],
		      (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_QUECVT+LCK$M_NODLCKWT), 0,
		      0, slaveStartIOGrantedAST, jbUnit, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void slaveStartIO(brnch$r_sndcmd_message* message):0") ;
	lib$signal(status,0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
}

static void
slaveStartIOGrantedAST(
    unsigned long jbUnit)
{
    unsigned long			status ;
    
    status = sys$enqw (0, LCK$K_NLMODE, &startIOLSB[jbUnit],
		       (LCK$M_SYSTEM+LCK$M_VALBLK+LCK$M_CONVERT+LCK$M_NODLCKWT), 0,
		       0, 0, 0, 0, 0, 0) ;
    if (!$VMS_STATUS_SUCCESS(status))
    {
	$DESCRIPTOR(xxx,"static void slaveStartIOGrantedAST(unsigned long jbUnit):1") ;
	lib$signal(status,0) ;
	lib$signal(BRNCH$_INTLOGERR, 1, &xxx) ;
    }

    return ;
} ;

static unsigned long
turnOnPhysicalDrive (
    unsigned long drive,
    unsigned long jbUnit)
{
    time_t				bPakakTime ;
    time_t				ePakakTime ;
    unsigned long			status ;

    if (acpgeo.brnch$l_pioneer604x != 0)
    {
	drive = jbUnit ;				/* The pioneer 604 box is weird.  It has 1 physical drive, but looks	*/
							/* like 6 physical drives.  We claim that it has 1 physical drive in	*/
							/* the acp geometry, but require it to have 6 physical drives in the	*/
							/* configuration table.  Because of this, we do a lot of faking for	*/
							/* this specific case.							*/
    }

    time(&bPakakTime) ;
    status = jpakak(drive, jbUnit) ;			/* Turn on the underlying physical device.				*/
    time(&ePakakTime) ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = jpakak(%d, %d) after %d seconds\n", status, drive, jbUnit, ePakakTime - bPakakTime) ;
    }

    if (!$VMS_STATUS_SUCCESS(status))
    {
	/*
	** The physical media has failed to enable itself.  Send a message to
	** the operator to log the fact and return with the error leaving the
	** device disconnected.
	*/

	char				jbUnitBuffer[256] ;
	struct dsc$descriptor_s		jbUnitDsc = {0, DSC$K_DTYPE_T, DSC$K_CLASS_S, jbUnitBuffer} ;
	struct dsc$descriptor_s		messageDsc = {0, DSC$K_DTYPE_T, DSC$K_CLASS_D, 0} ;

	jbUnitDsc.dsc$w_length =
	    sprintf (jbUnitBuffer, "%0.*s%d:", cte.brnch$t_jbDevice.dsc$w_length, cte.brnch$t_jbDevice.dsc$a_pointer, jbUnit) ;

	brnch_requestMessage (
	    &messageDsc,
	    0,
	    BRNCH$_PAKACK,
	    4,
	    &jbUnitDsc,
	    &cte.brnch$a_driveNames[drive],
	    status,
	    ePakakTime - bPakakTime) ;
	
	sys$sndopr (&messageDsc, 0) ;

	str$free1_dx (&messageDsc) ;
	return status ;
    }
    
    status = brnch_connectJBUnitToDrive(jbUnit, drive, SS$_NORMAL) ;
							/* Connect the JB unit to the physical device.				*/
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_connectJBUnitToDrive(%d, %d, %08x)\n", status, jbUnit, drive, SS$_NORMAL) ;
    }

    return status ;    
} ;

static void
turnOnPhysicalDriveFailed (
    unsigned long drive,
    unsigned long jbUnit)
{

    /*
    ** The enable of the physical drive has failed.  At this point, the
    ** following conditions obtain:
    **
    **	1. The PHYSICAL device may be ON the BUSY list.
    **	2. The PHYSICAL device is OFF the AVAILABLE list.
    **	3. The JB device is OFF the PENDING list.
    **	4. The PHYSICAL device has NOT been connected.
    **  5. The PHYSICAL and JB devices have been linked together.
    */
    
    struct brnch$r_inventory_table	diskRecord ;
    unsigned long			status ;

    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("Entering: turnOnPhysicalDriveFailed(%d, %d)\n", drive, jbUnit) ;
    }

    status = brnch__readIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
    if ($VMS_STATUS_SUCCESS(status))
    {
	diskRecord.brnch$r_overlay.brnch$l_contents[drive] = -1 ;
	brnch__updateIT(&diskRecord, BRNCH_K_IT_DRIVE_CONTENTS) ;
    }

    acpgeo.brnch$a_drives[drive].brnch$l_jbUnit = -1 ;	/* Break the linkages between JB and PHYSICAL device.			*/
    acpgeo.brnch$a_jbUnits[jbUnit].brnch$l_drive = -1 ;

    status = removeFromBusy(drive) ;

    if (status == 0)
    {
	brnch_removeFromAvailable(drive) ;
    } ;
    
    brnch_addToPending(jbUnit) ;			/* Restart the I/O request, the drive has been forgotten by VB.		*/

    acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onDisabled = 1 ;

    if (acpgeo.brnch$l_pioneer604x == 0)
    {
	/*
	** Empty the drive, just to make sure this is possible.  Technically, we
	** should check to make sure the unload works and, if it doesn't,
	** disable all the jb units associated with the media we couldn't get
	** out of the physical drive.
	*/
	
	brnch_unloadPhysicalDevice(drive) ;

	status = brnch_driveToVolume(drive, jbUnit) ;
    }

    status = 0 ;

    for (drive = 0;  drive < acpgeo.brnch$l_numDrives;  drive++)
    {
	if (!acpgeo.brnch$a_drives[drive].brnch$r_flags.brnch$v_onDisabled)
	{
	    status++ ;
	}
    }

    if (status == 0)
    {

	/*
	** There are no more drives in service.  Send a notification to the
	** operator and hose the server.
	*/
	
	struct dsc$descriptor_s		messageDsc = {0, DSC$K_DTYPE_T, DSC$K_CLASS_D, 0} ;

	brnch_requestMessage (
	    &messageDsc,
	    0,
	    BRNCH$_NODRIVE,
	    1,
	    &cte.brnch$t_name) ;
	
	sys$sndopr (&messageDsc, 0) ;

	str$free1_dx (&messageDsc) ;
	lib$signal (BRNCH$_NODRIVE, 1, &cte.brnch$t_name) ;
    }

    /*
    ** Restart the pending list scan which SHOULD get everything running again.
    ** Note that we SHOULD check to make sure that there are SOME drive
    ** available and if they aren't, to just give up the ghost more or less
    ** cleanly.
    */
    
    status = brnch_dispatchRoutine(masterProcessPendingList,-1) ;
    if (brnch_isTraceArbitrationOn())
    {
	brnch__logit("%08x = brnch_dispatchRoutine(masterProcessPendingList, -1)\n", status) ;
	brnch__logit("Exiting: turnOnPhysicalDriveFailed(%d, %d)\n", drive, jbUnit) ;
    }

    return ;
} ;
